from ..utils.custom_notation import CustomNotation


class Counter:
    def __init__(self):
        self.count = CustomNotation('0')

    def increment(self):
        self.count = CustomNotation(
            self.count.convert_from_decimal_notation(
                self.count.convert_to_decimal_notation() + 1
            )
        )

    def nullify(self):
        self.count = CustomNotation('0')



from datetime import datetime

from .abstract_storage import AbstractCallbackStorage
from ..utils.counter import Counter


class BuiltinCallbackStorage(AbstractCallbackStorage):

    def __init__(self):
        self._storage = dict()
        self._counter = Counter()
        self._pointers = list()

    def add_callback(self, callback: str):
        if not self._check_callback_existing(callback):
            abbreviation = self._pointers.pop() if len(self._pointers) > 0 else self._counter.count
            self._storage[callback] = dict()
            self._storage[callback]['abbreviation'] = abbreviation
            self._storage[callback]['created_at'] = datetime.now()
            self._storage[callback]['updated_at'] = datetime.now()
            self._storage[callback]['use_counter'] = 1
            self._counter.increment()
        else:
            self._storage[callback]['updated_at'] = datetime.now()
            self._storage[callback]['use_counter'] += 1

    def remove_callback(self, callback: str):
        if self._check_callback_existing(callback):
            self._storage[callback]['use_counter'] -= 1
            if self._storage[callback]['use_counter'] <= 0:
                self._pointers.append(self._storage.pop(callback))

    def get_callback(self, abbreviation: str) -> str:
        for key, value in self._storage.items():
            if value['abbreviation'] == abbreviation:
                return key

    def get_abbreviation(self, callback: str) -> str:
        return self._storage.get(callback)['abbreviation']

    def _check_callback_existing(self, callback: str) -> bool:
        return callback in self._storage

    def on_callbacks_exceeding_time_limit(self, cache_time_limit: int):
        for callback in self._storage:
            callback_time = datetime.now().timestamp() - self._storage[callback]['updated_at'].timesatmp()
            if callback_time > cache_time_limit:
                self._pointers.append(self._storage.pop(callback))

    def clear(self):
        self._storage = dict()
        self._pointers = list()
        self._counter.nullify()



from aiogram.utils.callback_data import CallbackData


class CallbackStackFactory(CallbackData):
    def __init__(self, prefix, *parts):
        super().__init__(prefix, *parts)
        self._part_names = *self._part_names, 'previous'



from abc import ABC, abstractmethod

from ..utils.singlton import AbstractSingleton


class AbstractCallbackStorage(ABC, metaclass=AbstractSingleton):

    @abstractmethod
    def add_callback(self, callback: str):
        pass

    @abstractmethod
    def remove_callback(self, callback: str):
        pass

    @abstractmethod
    def clear(self):
        pass



from contextlib import suppress
from enum import Enum

from .builtin_storage import BuiltinCallbackStorage

with suppress(ImportError):
    from .sqlite_storage import SQLiteCallbackStorage
with suppress(ImportError):
    from .redis_storage import RedisCallbackStorage


class StorageEnum(Enum):
    builtin = BuiltinCallbackStorage
    with suppress(NameError):
        sqlite = SQLiteCallbackStorage
    with suppress(NameError):
        redis = RedisCallbackStorage


class Storages:

    def __init__(self):
        self.storages = StorageEnum

    def get_storage(self, config):
        return self.storages[config['storage']].value()

    def add_storage(self, name, value):
        self.storages = Enum(
            'Storages',
            [(storage.name, storage.value) for storage in self.storages] + [(name, value)]
        )



from ..utils.callback_stack_factory import CallbackStackFactory


class CallbackStack:
    def __init__(self, callback_data: dict, callback_factory: CallbackStackFactory):
        self.separator = '¦'
        self.callback_factory = callback_factory
        self.callback_data = callback_data
        self.validate_callback_data()

    def validate_callback_data(self) -> None:
        if '@' in self.callback_data:
            self.callback_data.pop('@')

    def generate_previous_callback_data(self) -> list:
        return list(filter(None, self.callback_data.pop('previous').split(self.separator)))

    def get_complete_string_callback_data(
            self,
            callback_data: dict,
            callback_factory: CallbackStackFactory,
            previous_callback_data: list) -> str:

        try:
            return callback_factory.new(
                **callback_data,
                previous=self.get_string_previous_callback_data(previous_callback_data)
            )
        except ValueError:
            if not previous_callback_data:
                raise ValueError
            self.get_complete_string_callback_data(
                callback_data,
                callback_factory,
                previous_callback_data[1:]
            )

    def get_string_previous_callback_data(self, previous_callback_data: list) -> str:
        return self.separator.join(previous_callback_data)

    def get_string_callback_data(self) -> str:
        return self.callback_factory.new(
            **self.callback_data,
            previous=str())
